#include "midi.h"
#include <stdio.h>

#define MAXINT		0x7fffffff

struct Dest : public vector<unsigned short> {
	void push_note(int voice, int note, int velo) {
		vector<unsigned short>::push_back(voice << 11 | velo >> 3 << 7 | note);
	}
	void push_wait(int wait) {
		vector<unsigned short>::push_back(-wait);
	}
};

static Dest dest;

class Voice {
public:
	enum { N = 12 };
	Voice() {
		channel = note = velo = req = 0;
		elapse = MAXINT;
	}
	static void NoteOn(int channel, int note, int velo) {
		int max = 0;
		Voice *p, *maxp = voice;
		for (p = voice; p < voice + n; p++) {
			if (p->elapse == MAXINT) {
				maxp = p;
				break;
			}
			int v = 128 - p->velo;
			if (max < v) max = v, maxp = p;
		}
		if (maxp->elapse < MAXINT) maxp->req = 0, offcount++;
		if (maxvoice < maxp - voice) maxvoice = maxp - voice;
		maxp->On(channel, note, velo);
	}
	static void NoteOff(int channel, int note) {
		Voice *p;
		for (p = voice; p < voice + n && 
			(!(p->channel & 1 << channel) || p->note != note); p++) 
			;
		if (p < voice + n && !(p->channel &= ~(1 << channel))) p->Off();
	}
	static void Dump(int duration) {
		Voice *p;
		for (p = voice; p < voice + n; p++) {
			if (p->req) {
				p->req = 0;
				dest.push_note(p - voice, p->note, p->velo);
			}
			if (p->elapse < MAXINT) p->elapse += duration;
		}
	}
	static int OffCount() { return offcount; }
	static int MaxVoice() { return maxvoice + 1; }
	static void SetN(int v) {
		if (v < 1) v = 1;
		else if (v > N) v = N;
		n = v;
	}
private:
	void On(int _channel, int _note, int _velo) {
		channel = 1 << _channel;
		note = _note;
		velo = _velo;
		elapse = 0;
		req = 1;
	}
	void Off() {
		req = 1;
		channel = note = velo = 0;
		elapse = MAXINT;
	}
	int channel, note, velo, elapse, req;
	static Voice voice[N];
	static int offcount, maxvoice, n;
};

Voice Voice::voice[Voice::N];
int Voice::offcount, Voice::maxvoice, Voice::n = Voice::N;
static int enable = 0xfdff, started, base, fs = 44100;

static void note(int channel, int note, int velo) {
	if (enable & 1 << channel) {
		note -= base;
		if (velo) Voice::NoteOn(channel, note, velo);
		else Voice::NoteOff(channel, note);
		started = 1;
	}
}

static void wait(int duration) {
	if (started) {
		duration = int(duration / (1e6 / fs) + 0.5);
		Voice::Dump(duration);
		while (duration > 0x8000) {
			dest.push_wait(0x8000);
			duration -= 0x8000;
		}
		dest.push_wait(duration);
	}
}

int main(int argc, char *argv[]) {
	int v;
	while (argc > 2 && argv[1][0] == '-') {
		argc--;
		argv++;
		switch (argv[0][1]) {
			case 'e':
			sscanf(argv[1], "%x", &enable);
			argc--;
			argv++;
			break;
			case 'f':
			fs = 33113;
			break;
			case 'n':
			sscanf(argv[1], "%d", &v);
			argc--;
			argv++;
			Voice::SetN(v);
			break;
			case 't':
			sscanf(argv[1], "%d", &base);
			base = -base;
			argc--;
			argv++;
			break;
		}
	}
	if (argc != 2) {
		fprintf(stderr, "Usage: mid2pic [options] <filename>\noptions: [-e <enable channel>] [-n <# of voice>] [-t <transpose>] [-f]\n");
		return 1;
	}
	Midi midi;
	if (midi.Parse(argv[1])) return 1;
	int i, min, max;
	midi.GetNoteMinMax(min, max);
	if (min - base < 0) base += min - base;
	dest.push_wait(1);
	dest.push_wait(0x8000);
	midi.Dump(note, wait);
	for (i = 0; i < Voice::N; i++) dest.push_note(i, 0, 0);
	printf("MaxVoice: %d  Note: %d-%d\n", Voice::MaxVoice(), min - base, max - base);
	if (base) printf("Transpose: %d\n", -base);
	if (Voice::OffCount()) printf("OffCount: %d\n", Voice::OffCount());
	FILE *fo = fopen("orgel.dat", "wb");
	if (fo) {
		fwrite(&dest.front(), 2, dest.size(), fo);
		fclose(fo);
		printf("%d bytes\n", 2 * (int)dest.size());
	}
	fo = fopen("orgel.hex", "w");
	if (fo) {
		unsigned short *p = &dest.front();
		int i, d0, d1, t = 0, n = dest.size();
		for (i = 0; i < n; i++) {
			if (!(i & 7)) fprintf(fo, ":%02X%04X00", n - i < 8 ? n - i << 1 : 16, i << 1);
			t += d0 = p[i] >> 8;
			t += d1 = p[i] & 0xff;
			fprintf(fo, "%02X%02X", d0, d1);
			if ((i & 7) == 7) fprintf(fo, "%02X\r\n", -t & 0xff);
		}
		fclose(fo);
	}
	fo = fopen("orgel.asm", "w");
	if (fo) {
		unsigned short *p = &dest.front();
		int i, n = dest.size();
		fprintf(fo, "orgelcode:\r\n\tmovwf\tPCL\r\n");
		for (i = 0; i < 63; i++) 
			fprintf(fo, "\tretlw\t0x%02x\r\n\tretlw\t0x%02x\r\n", p[i] & 0xff, p[i] >> 8);
		fclose(fo);
	}
	return 0;
}

