#include "midi.h"
#include <stdio.h>

static void *readall(FILE *fi, long *oSize) {
	size_t size;
	fseek(fi, 0, SEEK_END);
	size = ftell(fi);
	rewind(fi);
	void *buf = malloc(size);
	if (buf) {
		fread(buf, size, 1, fi);
		if (oSize) *oSize = size;
	}
	return buf;
}

int Midi::Parse(char *filename) {
	FILE *fi;
	if (!(fi = fopen(filename, "rb"))) {
		fprintf(stderr, "%s cannot opened.\n", filename);
		return 1;
	}
	long filelen;
	buf = (u8 *)readall(fi, &filelen);
	fclose(fi);
	bufp = buf + 12;
	timebase = v2();
	while (bufp < buf + filelen) {
		bufp += 4;
		int len = v4();
		u8 *limit = bufp + len;
		int time = 0, status = 0;
		while (bufp < limit) {
			time += vv();
			int len, a, eventtype;
			if (*bufp & 0x80) status = *bufp++;
			switch (status) {
				case 0x80: case 0x81: case 0x82: case 0x83:
				case 0x84: case 0x85: case 0x86: case 0x87:
				case 0x88: case 0x89: case 0x8a: case 0x8b:
				case 0x8c: case 0x8d: case 0x8e: case 0x8f:
				// note off
				a = v1();
				v1();
				data.push_back(DataElm::NOTE, time, status & 0xf, a, 0);
				break;
				case 0x90: case 0x91: case 0x92: case 0x93:
				case 0x94: case 0x95: case 0x96: case 0x97:
				case 0x98: case 0x99: case 0x9a: case 0x9b:
				case 0x9c: case 0x9d: case 0x9e: case 0x9f:
				// note on
				a = v1();
				if (notemax < a) notemax = a;
				if (notemin > a) notemin = a;
				data.push_back(DataElm::NOTE, time, status & 0xf, a, volume[status & 0xf] * expression[status & 0xf] * v1() >> 14);
				break;
				case 0xa0: case 0xa1: case 0xa2: case 0xa3:
				case 0xa4: case 0xa5: case 0xa6: case 0xa7:
				case 0xa8: case 0xa9: case 0xaa: case 0xab:
				case 0xac: case 0xad: case 0xae: case 0xaf:
				// key pressure
				bufp += 2;
				break;
				case 0xb0: case 0xb1: case 0xb2: case 0xb3:
				case 0xb4: case 0xb5: case 0xb6: case 0xb7:
				case 0xb8: case 0xb9: case 0xba: case 0xbb:
				case 0xbc: case 0xbd: case 0xbe: case 0xbf:
				// control change
				a = v1();
				switch (a) {
					case 0x07: case 0x27:
					volume[status & 0xf] = v1();
					break;
					case 0x0b: case 0x2b:
					expression[status & 0xf] = v1();
					break;
					default:
					bufp++;
					break;
				}
				break;
				case 0xc0: case 0xc1: case 0xc2: case 0xc3:
				case 0xc4: case 0xc5: case 0xc6: case 0xc7:
				case 0xc8: case 0xc9: case 0xca: case 0xcb:
				case 0xcc: case 0xcd: case 0xce: case 0xcf:
				// program change
				bufp++;
				break;
				case 0xd0: case 0xd1: case 0xd2: case 0xd3:
				case 0xd4: case 0xd5: case 0xd6: case 0xd7:
				case 0xd8: case 0xd9: case 0xda: case 0xdb:
				case 0xdc: case 0xdd: case 0xde: case 0xdf:
				// channel pressure
				bufp++;
				break;
				case 0xe0: case 0xe1: case 0xe2: case 0xe3:
				case 0xe4: case 0xe5: case 0xe6: case 0xe7:
				case 0xe8: case 0xe9: case 0xea: case 0xeb:
				case 0xec: case 0xed: case 0xee: case 0xef:
				// pitch bend change
				bufp += 2;
				break;
				case 0xf0: case 0xf7:
				bufp += vv();
				break;
				case 0xff:
				eventtype = *bufp++;
				len = vv();
				switch (eventtype) {
					case 0x51:
					data.push_back(DataElm::TEMPO, time, v3(), 0, 0);
					break;
					default:
					bufp += len;
					break;
				}
				break;
				default:
				fprintf(stderr, "Illegal data ofs=%x\n", bufp - buf);
				goto ERROR;
			}
		}
	}
ERROR:
	free(buf);
	buf = 0;
	return 0;
}

int Midi::comp(const void *a, const void *b) {
	int ta = ((DataElm *)a)->time;
	int tb = ((DataElm *)b)->time;
	return ta < tb ? -1 : ta > tb ? 1 : 0;
}

void Midi::Dump(NoteCallback note, WaitCallback wait) {
	qsort(&data.front(), data.size(), sizeof(DataElm), comp);
	int last = 0, tempo = 0, delta;
	for (Data::iterator i = data.begin(); i != data.end(); i++) {
		switch (i->id) {
			case DataElm::TEMPO:
			tempo = i->v0;
			break;
			case DataElm::NOTE:
			delta = tempo * (i->time - last) / timebase;
			last = i->time;
			if (delta && wait) wait(delta);
			if (note) note(i->v0, i->v1, i->v2);
			break;
		}
	}
}

