#ifndef _MIDI_H_
#define _MIDI_H_

#include <vector.h>

typedef unsigned char u8;
typedef void (*NoteCallback)(int channel, int note, int velo);
typedef void (*WaitCallback)(int duration);

class Midi {
	struct DataElm {
		enum { NOTE, TEMPO };
		int id, time, v0, v1, v2;
	};
	struct Data : public vector<DataElm> {
		void push_back(int id, int time, int v0, int v1, int v2) {
			DataElm d = { id, time, v0, v1, v2 };
			vector<DataElm>::push_back(d);
		}
	};
public:
	Midi() {
		buf = bufp = 0;
		timebase = notemax = 0;
		notemin = 127;
		for (int i = 0; i < 16; i++) {
			volume[i] = 100;
			expression[i] = 127;
		}
	}
	virtual ~Midi() {
		free(buf);
		buf = 0;
	}
	int Parse(char *filename);
	void Dump(NoteCallback note, WaitCallback wait);
	void GetNoteMinMax(int &min, int &max) const {
		max = notemax;
		min = notemin;
	}
private:
	int v1() {
		return *bufp++;
	}
	int v2() {
		bufp += 2;
		return bufp[-2] << 8 | bufp[-1];
	}
	int v3() {
		bufp += 3;
		return bufp[-3] << 16 | bufp[-2] << 8 | bufp[-1];
	}
	int v4() {
		bufp += 4;
		return bufp[-4] << 24 | bufp[-3] << 16 | bufp[-2] << 8 | bufp[-1];
	}
	int vv() {
		int d, r = 0;
		do {
			d = *bufp++;
			r = r << 7 | d & 0x7f;
		} while (d & 0x80);
		return r;
	}
	static int comp(const void *a, const void *b);
	u8 *buf, *bufp;
	int timebase, notemax, notemin;
	u8 volume[16], expression[16];
	Data data;
};

#endif

